#!/usr/bin/python3

import copy
import os
import sys
import yaml

import editor

from jira.exceptions import JIRAError

from jirate.args import ComplicatedArgs, GenericArgs
from jirate.jboard import JiraProject, get_jira
from jirate.decor import md_print, pretty_date, color_string, hbar_under, hbar, hbar_over, nym, vsep_print, vseparator
from jirate.decor import pretty_print  # NOQA
from jirate.config import get_config
from jirate.jira_fields import apply_field_renderers, render_issue_fields, max_field_width


def move(args):
    if args.user:
        args.project.assign(args.src, args.user)
    if args.mine:
        args.project.assign(args.src, 'me')
    if args.project.move(args.src, args.target):
        print('Moved', args.src, 'to', args.target)
        return (0, False)
    return (1, False)


def close_issues(args):
    ret = 0
    for issue in args.target:
        if not args.project.close(issue):
            ret = 1
    return (ret, False)


def print_issues_by_state(issue_list, args=None):
    states = {}

    for issue in issue_list:
        cstatus = issue.raw['fields']['status']['name']
        if cstatus not in states:
            states[cstatus] = []
        states[cstatus].append(issue)

    for key in states:
        if args and args.status and nym(key) != nym(args.status):
            continue
        hbar_under(key)
        for issue in states[key]:
            print('  ', issue.key, end=' ')
            if args and args.labels:
                print_labels(issue.raw, prefix='')
            print(issue.raw['fields']['summary'])
        print()


def print_users(users):
    nsize = len('Name')
    ksize = len('User Name')
    msize = len('Email Address')

    for user in users:
        nsize = max(nsize, len(user.displayName))
        ksize = max(ksize, len(user.name))
        msize = max(msize, len(user.emailAddress))

    header = 'Name'.ljust(nsize) + '   ' + 'User Name'.ljust(ksize) + '   ' + 'Email Address'.ljust(msize)
    hbar_under(header)
    for user in users:
        vsep_print(None, user.displayName, nsize, user.name, ksize, user.emailAddress)


def search_jira(args):
    if args.user:
        users = args.project.search_users(args.user)
        if not users:
            print('No users match "f{args.user}"')
            return (1, False)
        print_users(users)
        return (0, False)

    named = args.named_search
    if not args.text and not named:
        named = 'default'
    if named:
        searches = args.project.get_user_data('searches')
        if named not in searches:
            print(f'No search configured: {named}')
            return (1, False)
        search_query = searches[named]
        ret = args.project.search_issues(search_query)
    else:
        search_query = ' '.join(args.text)
        if args.raw:
            ret = args.project.search_issues(search_query)
        else:
            ret = args.project.search(search_query)

    if not ret:
        return (127, False)
    print_issues_by_state(ret)
    hbar_over(str(len(ret)) + ' result(s)')
    return (0, False)


def list_issues(args):
    # check for verbose
    if args.mine:
        userid = 'me'
    elif args.unassigned:
        userid = 'none'
    elif args.user:
        userid = args.user
    else:
        userid = None

    issues = args.project.list(userid=userid)
    print_issues_by_state(issues, args)
    return (0, True)


def list_link_types(args):
    ltypes = args.project.link_types()
    namelen = len('Outward')
    for lt in ltypes:
        namelen = max(namelen, len(lt.outward))
    blen = vsep_print(None, 'Outward', namelen, 'Inward')
    hbar(blen)
    for lt in ltypes:
        vsep_print(None, lt.outward, namelen, lt.inward)
    return (0, True)


def list_states(args):
    states = args.project.states()
    namelen = 0
    for name in states:
        namelen = max(namelen, len(name))
    for name in states:
        vsep_print(None, name, namelen, states[name]['name'])
    return (0, False)


def list_issue_types(args):
    issue_types = args.project.issue_types
    for itype in issue_types:
        print('  ', itype.name)
    return (0, False)


def issue_fields(args):
    if args.issue:
        issue = args.project.issue(args.issue)
        if not issue:
            print('No such issue:', args.issue)
            return (1, False)
        fields = args.project.fields(issue.raw['key'])
    elif args.type:
        md = args.project.issue_metadata(args.type)
        if md:
            fields = md['fields']
        else:
            print(f'No metadata for {args.type}')
            return (1, False)

    # Remove things we set elsewhere
    for field in ('description', 'summary', 'assignee', 'issuelinks', 'comment'):
        if field in fields:
            del fields[field]

    # Remove things we don't support setting
    for field in ('issuetype', 'attachment', 'reporter', 'project'):
        if field in fields:
            del fields[field]

    display = False
    try:
        if args.operation:
            pass
    except AttributeError:
        display = True

    if display:
        nlen = 0
        for field in fields:
            if field.startswith('customfield_'):
                nlen = max(nlen, len(nym(fields[field]['name'])))
            else:
                nlen = max(nlen, len(nym(field)))
        for field in fields:
            if field.startswith('customfield_'):
                fname = nym(fields[field]['name'])
            else:
                fname = nym(field)
            fvalue = ''
            if 'allowedValues' in fields[field]:
                values = []
                for val in fields[field]['allowedValues']:
                    if 'archived' in val and val['archived']:
                        continue
                    if 'name' in val:
                        values.append(val['name'])
                    elif 'value' in val:
                        values.append(val['value'])
                    else:
                        values.append(val['id'])
                fvalue = ', '.join(values)
            vsep_print(' ', fname, nlen, fvalue)
        return (0, False)

    field = None
    for _field in fields:
        if args.name not in (field, fields[_field]['name'], nym(fields[_field]['name']), fields[_field]['fieldId'], nym(fields[_field]['fieldId'])):
            continue
        field = fields[_field]
        break

    if not field:
        key = issue.raw['key']
        print(f'No field like \'{args.name}\' field in {key}')
        return (1, False)

    ops = field['operations']
    if args.operation not in ops:
        print(f'Cannot perform {args.operation} on {args.issue}; try: {ops}')
        return (1, False)

    # Join stuff if it's not an array
    if 'schema' in field and field['schema']['type'] == 'array':
        start_val = args.values
    else:
        start_val = [' '.join(args.values)]

    # okay time to update them
    found = False
    send_val = []

    # Parse allowedValues and look for name, value, and ID, and their nyms
    if 'allowedValues' in field:
        # Validate that the name or value exists and create our array of IDs
        # corresponding to them.
        for val in start_val:
            found = False
            for av in field['allowedValues']:
                if 'archived' in av and av['archived']:
                    continue
                for key in ['name', 'value']:
                    if key not in av:
                        continue
                    if val not in (av[key], nym(av[key])):
                        continue
                    send_val.append({'id': av['id']})
                    found = True
                    break
                if found:
                    break
            if not found:
                print(f'Value {val} not allowed for {args.name}')
                return (1, False)

    # Start with our basic input otherwise; no validation done
    else:
        send_val = start_val

    update_args = []
    # If it's not an array, assume a string for now
    if 'schema' not in field or field['schema']['type'] != 'array':
        send_val = send_val[0]

    # Add and remove use a different format than 'set'.
    # There's also 'modify', but ... that one's even more complicated.
    if args.operation in ['add', 'remove']:
        update_args = {field['fieldId']: [{args.operation: val} for val in send_val]}
    else:
        update_args = {field['fieldId']: [{args.operation: send_val}]}
    args.project.update_issue(issue.raw['key'], **update_args)
    return (0, False)


def split_issue_text(text):
    lines = text.split('\n')
    name = lines[0]
    desc = ''
    if not len(name):
        return (None, None)
    lines.pop(0)
    while len(lines) and lines[0] == '':
        lines.pop(0)
    if len(lines):
        desc = '\n'.join(lines)
    return (name, desc)


def new_issue(args):
    desc = None

    if args.text:
        name = ' '.join(args.text)
    else:
        text = editor()
        name, desc = split_issue_text(text)
        if name is None:
            print('Canceled')
            return (1, False)

    issue = args.project.new(name, desc, issue_type=args.type)
    if args.quiet:
        print(issue.raw['key'])
    else:
        print_issue(args.project, issue, False)
    return (0, True)


# new_issue is way too easy. Let's make it *incredibly* complicated!
def create_issue(args):
    auto_fields = ['reporter']
    desc = None
    issuetype = args.type if args.type else 'Task'

    # Do sanity check before hitting the JIRA API
    if len(args.args) % 2 == 1:
        print('Incorrect number of arguments (not divisible by 2)')
        return (1, False)

    try:
        metadata = args.project.issue_metadata(args.type)
    except JIRAError as e:
        if 'text: Issue Does Not Exist' in str(e):
            print('The createmeta API does not exist on this JIRA instance.')
        else:
            print(e)
        return (1, False)

    if not metadata:
        print('Invalid issue type:', issuetype)
        print(f'Valid issue types for {args.project.project_name}:')
        list_issue_types(args)
        return (1, False)

    values = {}
    argv = copy.copy(args.args)
    while len(argv):
        key = argv.pop(0)
        value = argv.pop(0)
        values[key] = value

    # Bug - error checking isn't done until later, but we need a
    # summary or the below code blows up.  So, you might pop up
    # $EDITOR only to find out later that you used an invalid
    # issue type.
    if 'summary' not in values:
        text = editor()
        name, desc = split_issue_text(text)
        if name is None:
            print('Canceled')
            return (1, False)
        values['summary'] = name
        values['description'] = desc

    # Nope all the auto-populated things
    for field in auto_fields:
        if field in values:
            del values[field]

    commit_values = {}
    errors = 0

    issuetype = metadata['name']
    values['issuetype'] = issuetype
    values['project'] = args.project.project_name

    # resolve field names
    for field in metadata['fields']:
        if field in auto_fields:
            continue
        fieldname = metadata['fields'][field]['name']
        if field not in values and fieldname not in values and nym(fieldname) not in values:
            if metadata['fields'][field]['required']:
                print(f'Missing required field for {args.project.project_name}/{issuetype}: ' + nym(fieldname))
                errors = errors + 1
            continue
        if field in values:
            commit_values[field] = values[field]
            del values[field]
        if fieldname in values:
            commit_values[field] = values[fieldname]
            del values[fieldname]
        if nym(fieldname) in values:
            commit_values[field] = values[nym(fieldname)]
            del values[nym(fieldname)]

    if values:
        print('WARNING: Input fields is not empty:')
        pretty_print(values)

    if errors:
        print(f'{errors} errors; can\'t create issue')
        return (1, False)

    issue = args.project.create(**commit_values)
    if args.quiet:
        print(issue.raw['key'])
    else:
        print_issue(args.project, issue, False)
    return (0, True)


def create_from_template(args):
    # TODO: consider using Jira's bulk issue creation
    # TODO: support reading arbitrary fields from the template
    all_filed = []  # We keep issue keys here because we'll need to refresh anyway
    with open(args.template_file, 'r') as yaml_file:
        template = yaml.safe_load(yaml_file)

    for issue in template['issues']:
        filed = {}
        if 'description' not in issue:
            issue['description'] = ""
        parent = args.project.new(issue['summary'], description=issue['description'], issue_type=issue['issue_type'])
        filed['parent'] = parent.raw['key']

        if issue['subtasks']:
            filed['subtasks'] = []
            for subtask in issue['subtasks']:
                if 'description' not in subtask:
                    subtask['description'] = ""
                child = args.project.subtask(parent.raw['key'], subtask['summary'], subtask['description'])
                filed['subtasks'].append(child.raw['key'])
        all_filed.append(filed)

    # Need to refresh to that issues get re-fetched to include subtasks
    # TODO: Have subtask() update parent issues in _config['issue_map']
    args.project.delete_issue_map()
    for filed in all_filed:
        if args.quiet:
            if 'subtasks' in filed:
                print(filed['parent'] + ': ' + ', '.join(filed['subtasks']))
            else:
                print(filed['parent'])
        else:
            print_issue(args.project, args.project.issue(filed['parent']), False)
    return (0, True)


def new_subtask(args):
    desc = None
    parent_issue = args.project.issue(args.issue_id)

    if args.text:
        name = ' '.join(args.text)
    else:
        text = editor()
        name, desc = split_issue_text(text)
        if name is None:
            print('Canceled')
            return (1, False)

    issue = args.project.subtask(parent_issue.raw['key'], name, desc)
    if args.quiet:
        print(issue.raw['key'])
    else:
        print_issue(args.project, issue, False)
    return (0, True)


def link_issues(args):
    left_issue = args.issue_left
    right_issue = args.issue_right
    link_name = ' '.join(args.text)

    args.project.link(left_issue, right_issue, link_name)
    return (0, True)


def unlink_issues(args):
    args.project.unlink(args.issue_left, args.issue_right)
    return (0, True)


def link_url(args):
    issue = args.issue
    url = args.url
    text = ' '.join(args.text)
    args.project.attach(issue, url, text)
    return (0, True)


def comment(args):
    issue_id = args.issue

    if args.remove:
        comment_id = args.remove
        comment = args.project.get_comment(issue_id, comment_id)
        comment.delete()
        return (0, False)

    if args.edit:
        comment_id = args.edit
        comment = args.project.get_comment(issue_id, comment_id)
        if args.text:
            new_text = ' '.join(args.text)
        else:
            new_text = editor(comment.body)
            if not new_text:
                print('Canceled')
                return (0, False)
        if comment.body != new_text:
            comment.update(body=new_text)
        else:
            print('No changes')
        return (0, False)

    if args.text:
        text = ' '.join(args.text)
    else:
        text = editor()

    if not len(text):
        print('Canceled')
        return (0, False)

    args.project.comment(issue_id, text)
    return (0, False)


def refresh(args):
    args.project.refresh()
    args.project.index_issues()
    return (0, True)


def display_comment(action, verbose):
    print(pretty_date(action['updated']), '•', action['updateAuthor']['emailAddress'], '-', action['updateAuthor']['displayName'], '• ID:', action['id'])
    md_print(action['body'])
    print()


def display_attachment(attachment, verbose):
    print('  ' + attachment['name'])
    if verbose:
        print('    ID:', attachment['id'])
    if attachment['isUpload']:
        if attachment['filename'] != attachment['name']:
            print('    Filename:', attachment['filename'])
    else:
        if attachment['url'] != attachment['name']:
            print('    URL:', attachment['url'])


def print_labels(issue, prefix='Labels: '):
    if 'labels' in issue and len(issue['labels']):
        print(prefix, end='')
        for label in issue['labels']:
            print(label, end=' ')
        print()


def print_issue_links(issue):
    hbar_under('Issue Links')
    # pass 1: Get the lengths so we can draw separators
    sep = f' {vseparator} '
    lsize = 0
    rsize = 0
    for link in issue['issuelinks']:
        if 'outwardIssue' in link:
            text = link['type']['outward'] + ' ' + link['outwardIssue']['key']
            status = link['outwardIssue']['fields']['status']['name']
        elif 'inwardIssue' in link:
            text = link['type']['inward'] + ' ' + link['inwardIssue']['key']
            status = link['inwardIssue']['fields']['status']['name']

        if len(text) > lsize:
            lsize = len(text)
        if len(status) > rsize:
            rsize = len(status)
    # pass 2: print the stuff
    for link in issue['issuelinks']:
        if 'outwardIssue' in link:
            text = link['type']['outward'] + ' ' + link['outwardIssue']['key']
            status = link['outwardIssue']['fields']['status']
            desc = link['outwardIssue']['fields']['summary']
        elif 'inwardIssue' in link:
            text = link['type']['inward'] + ' ' + link['inwardIssue']['key']
            status = link['inwardIssue']['fields']['status']
            desc = link['inwardIssue']['fields']['summary']
        # color_string throws off length calculations
        vsep_print(' ', text.ljust(lsize) + sep + color_string(status['name'].ljust(rsize), status['statusCategory']['colorName']), lsize + rsize + 3, desc)
    print()


def print_remote_links(links):
    hbar_under('External Links')

    # pass 1: Get the lengths so we can draw separators
    lsize = 0
    rsize = 0
    for link in links:
        text = link.raw['object']['title']
        lid = str(link.raw['id'])
        if len(lid) > lsize:
            lsize = len(lid)
        if len(text) > rsize:
            rsize = len(text)
    # pass 2: print the stuff
    for link in links:
        # color_string throws off length calculations
        text = link.raw['object']['title']
        lid = str(link.raw['id'])
        url = link.raw['object']['url']
        vsep_print(' ', lid.ljust(lsize), lsize, text.ljust(rsize), rsize, url)
    print()


# Dict from search or subtask list
def _print_issue_list(header, issues):
    if not issues:
        return
    hbar_under(header)
    # pass 1: Get the lengths so we can draw separators
    sep = f' {vseparator} '
    lsize = 0
    rsize = 0
    for task in issues:
        if isinstance(task, str):
            task = issues[task]
        try:
            task_key = task.key
            status = task.raw['fields']['status']['name']
        except AttributeError:
            task_key = task['key']
            status = task['fields']['status']['name']
        if len(task_key) > lsize:
            lsize = len(task_key)
        if len(status) > rsize:
            rsize = len(status)
    # pass 2: print the stuff
    for task in issues:
        if isinstance(task, str):
            task = issues[task]
        try:
            task_key = task.key
            status = task.raw['fields']['status']
            summary = task.raw['fields']['summary']
        except AttributeError:
            task_key = task['key']
            status = task['fields']['status']
            summary = task['fields']['summary']
        # color_string throws off length calculations
        vsep_print(' ', task_key.ljust(lsize) + sep + color_string(status['name'].ljust(rsize), status['statusCategory']['colorName']), lsize + rsize + 3, summary)
    print()


def print_subtasks(issue):
    _print_issue_list('Sub-tasks', issue['subtasks'])


def print_issue(project, issue_obj, verbose=False, no_comments=False):
    issue = issue_obj.raw['fields']
    lsize = max(len(issue_obj.raw['key']), max_field_width(issue, verbose, project.allow_code))

    vsep_print(' ', issue_obj.raw['key'], lsize, issue['summary'])
    render_issue_fields(issue, verbose, project.allow_code, lsize)

    if verbose:
        vsep_print(' ', 'ID', lsize, issue_obj.raw['id'])
        vsep_print(None, 'URL', lsize, issue_obj.permalink())
        trans = project.transitions(issue_obj.raw['key'])
        if trans:
            vsep_print(' ', 'Next States', lsize, [tr['name'] for tr in trans.values()])
        else:
            vsep_print(None, 'Next States', lsize, 'No valid transitions; cannot alter status')

    print()
    if issue['description']:
        md_print(issue['description'])
        print()

    if 'issuelinks' in issue and len(issue['issuelinks']):
        print_issue_links(issue)

    # Don't print external links unless in verbose mode since it's another API call?
    if verbose:
        links = project.remote_links(issue_obj)
        if links:
            print_remote_links(links)

    if 'subtasks' in issue and len(issue['subtasks']):
        print_subtasks(issue)

    if issue['issuetype']['name'] == 'Epic':
        ret = project.search_issues('"Epic Link" = "' + issue_obj.raw['key'] + '"')
        _print_issue_list('Issues in Epic', ret)

    if no_comments:
        return
    if issue['comment']['comments']:
        hbar_under('Comments')

        for cmt in issue['comment']['comments']:
            display_comment(cmt, verbose)


def cat(args):
    issues = []
    for issue_idx in args.issue_id:
        issue = args.project.issue(issue_idx, True)
        if not issue:
            print('No such issue:', issue_idx)
            return (127, False)
        issues.append(issue)

    for issue in issues:
        print_issue(args.project, issue, args.verbose, args.no_comments)
    return (0, False)


def join_issue_text(name, desc):
    if desc:
        return name + '\n\n' + desc
    return name + '\n\n'


def edit_issue(args):
    issue_idx = args.issue

    issue_obj = args.project.issue(issue_idx)
    issue = issue_obj.raw['fields']
    issue_text = join_issue_text(issue['summary'], issue['description'])
    if args.text:
        new_text = ' '.join(args.text)
    else:
        new_text = editor(issue_text)
    if not new_text:
        print('Canceled')
        return (0, False)
    name, desc = split_issue_text(new_text)
    update_args = {}
    if issue['summary'] != name and issue['summary']:
        update_args['summary'] = name
    if issue['description'] != desc:
        update_args['description'] = desc
    if update_args != {}:
        args.project.update_issue(issue_idx, **update_args)
    else:
        print('No changes')
    return (0, False)


def view_issue(args):
    issue_id = args.issue_id
    issue = args.project.issue(issue_id)
    if not issue:
        return (127, False)
    os.system('xdg-open ' + issue.permalink())
    return (0, False)


def assign_issue(args):
    args.project.assign(args.issue_id, args.user)
    return (0, False)


def unassign_issue(args):
    args.project.assign(args.issue_id, 'none')
    return (0, False)


def user_info(args):
    user_info = [GenericArgs(args.project.user)]
    print_users(user_info)
    return (0, False)


def get_project(project=None, config=None, config_file=None):
    # project: Project key
    # config: dict / pre-read JSON data
    if not config:
        config = get_config(config_file)
    allow_code = False

    if 'jira' not in config:
        print('No JIRA configuration available')
        return None
    jconfig = config['jira']

    if not project:
        if 'default_project' in jconfig and not project:
            project = jconfig['default_project']
        else:
            print('No JIRA project specified')
            return None

    # Allows users to represent custom fields in output.
    # Not recommended to enable.
    if 'here_there_be_dragons' in jconfig:
        if jconfig['here_there_be_dragons'] is True:
            allow_code = True

    if not project:
        # Not sure why I used an array here
        project = jconfig['default_project']
    if 'proxies' not in jconfig:
        jconfig['proxies'] = {"http": "", "https": ""}

    jira = get_jira(jconfig)
    proj = JiraProject(jira, project, readonly=False, allow_code=allow_code, simplify=True)
    if 'searches' in jconfig:
        proj.set_user_data('searches', jconfig['searches'])
    if 'custom_fields' in jconfig:
        proj.custom_fields = copy.deepcopy(jconfig['custom_fields'])
        apply_field_renderers(proj.custom_fields)

    return proj


def create_parser():
    parser = ComplicatedArgs()

    parser.add_argument('-p', '--project', help='Use this JIRA project instead of default', default=None, type=str.upper)

    cmd = parser.command('whoami', help='Display current user information', handler=user_info)

    cmd = parser.command('ls', help='List issue(s)', handler=list_issues)
    cmd.add_argument('-m', '--mine', action='store_true', help='Display only issues assigned to me.')
    cmd.add_argument('-U', '--unassigned', action='store_true', help='Display only issues with no assignee.')
    cmd.add_argument('-u', '--user', help='Display only issues assigned to the specific user.')
    cmd.add_argument('-l', '--labels', action='store_true', help='Display issue labels.')
    cmd.add_argument('status', nargs='?', default=None, help='Restrict to issues in this state')

    cmd = parser.command('search', help='Search issue(s)/user(s) with matching text', handler=search_jira)
    cmd.add_argument('-u', '--user', help='Search for user(s) (max)')
    cmd.add_argument('-n', '--named-search', help='Perform preconfigured named search for issues')
    cmd.add_argument('-r', '--raw', action='store_true', help='Perform raw JQL query')
    cmd.add_argument('text', nargs='*', help='Search text')

    cmd = parser.command('cat', help='Print issue(s)', handler=cat)
    cmd.add_argument('-v', '--verbose', action='store_true', help='Verbose output')
    cmd.add_argument('-N', '--no-comments', action='store_true', default=False, help='Skip comments')
    cmd.add_argument('issue_id', nargs='+', help='Target issue(s)', type=str.upper)

    cmd = parser.command('view', help='Display issue in browser', handler=view_issue)
    cmd.add_argument('issue_id', help='Target issue', type=str.upper)

    parser.command('ll', help='List states available to project', handler=list_states)
    parser.command('lt', help='List issue types available to project', handler=list_issue_types)
    parser.command('link-types', help='Display link types', handler=list_link_types)

    cmd = parser.command('assign', help='Assign issue', handler=assign_issue)
    cmd.add_argument('issue_id', help='Target issue', type=str.upper)
    cmd.add_argument('user', help='Target assignee')
    # cmd.add_argument('users', help='First is assignee; rest are watchers (if none, assign to self)', nargs='*')

    cmd = parser.command('unassign', help='Remove assignee from issue', handler=unassign_issue)
    cmd.add_argument('issue_id', help='Target issue', type=str.upper)

    cmd = parser.command('mv', help='Move issue(s) to new state', handler=move)
    cmd.add_argument('-m', '--mine', action='store_true', help='Also assign to myself')
    cmd.add_argument('-u', '--user', help='Also assign to user')
    cmd.add_argument('src', metavar='issue', nargs='+', help='Issue key(s)')
    cmd.add_argument('target', help='Target state')

    cmd = parser.command('new', help='Create a new issue', handler=new_issue)
    cmd.add_argument('-t', '--type', default='task', help='Issue type (project-dependent)')
    cmd.add_argument('-q', '--quiet', default=False, help='Only print new issue ID after creation (for scripting)', action='store_true')
    cmd.add_argument('text', nargs='*', help='Issue summary')

    cmd = parser.command('create', help='Create a new issue (advanced)', handler=create_issue)
    cmd.add_argument('-t', '--type', default='task', help='Issue type (project-dependent)')
    cmd.add_argument('-q', '--quiet', default=False, help='Only print new issue ID after creation (for scripting)', action='store_true')
    cmd.add_argument('args', nargs='*', help='field1 "value1" field2 "value2" ... fieldN "valueN"')

    cmd = parser.command('subtask', help='Create a new subtask', handler=new_subtask)
    cmd.add_argument('-q', '--quiet', default=False, help='Only print subtask ID after creation (for scripting)', action='store_true')
    cmd.add_argument('issue_id', help='Parent issue', type=str.upper)
    cmd.add_argument('text', nargs='*', help='Subtask summary')

    cmd = parser.command('link', help='Create link between two issues', handler=link_issues)
    cmd.add_argument('issue_left', help='First issue', type=str.upper)
    cmd.add_argument('text', nargs='+', help='Link text')
    cmd.add_argument('issue_right', help='Second issue', type=str.upper)

    cmd = parser.command('attach', help='Attach a web link to an issue', handler=link_url)
    cmd.add_argument('issue', help='Issue', type=str.upper)
    cmd.add_argument('url', help='URL to attach')
    cmd.add_argument('text', nargs='+', help='URL Description')

    cmd = parser.command('unlink', help='Remove link(s) between issues or an external link', handler=unlink_issues)
    cmd.add_argument('issue_left', help='First issue', type=str.upper)
    cmd.add_argument('issue_right', help='Second issue (or external link ID)', type=str.upper)

    cmd = parser.command('comment', help='Comment (or remove) on an issue', handler=comment)
    cmd.add_argument('-e', '--edit', help='Comment ID to edit')
    cmd.add_argument('-r', '--remove', help='Comment ID to remove')
    cmd.add_argument('issue', help='Issue to operate on')
    cmd.add_argument('text', nargs='*', help='Comment text')

    cmd = parser.command('edit', help='Edit issue description or summary', handler=edit_issue)
    cmd.add_argument('issue', help='Issue')
    cmd.add_argument('text', nargs='*', help='New text')

    cmd = parser.command('field', help='Update field values for an issue', handler=issue_fields)
    cmd.add_argument('issue', help='Issue')
    cmd.add_argument('operation', help='Operation', choices=['add', 'set', 'remove'])
    cmd.add_argument('name', help='Name of field to update')
    cmd.add_argument('values', help='Value(s) to update', nargs='*')

    cmd = parser.command('fields', help='List fields (and allowed values, when applicable)', handler=issue_fields)
    cmd.add_argument('-t', '--type', default=None, help='Fields available at creation time for the specified type')
    cmd.add_argument('issue', help='Existing Issue (more fields available here)', nargs='?')

    cmd = parser.command('close', help='Move issue(s) to closed/done/resolved', handler=close_issues)
    cmd.add_argument('target', nargs='+', help='Target issue(s)')

    cmd = parser.command('template', help='Create issue from YAML template', handler=create_from_template)
    cmd.add_argument('template_file', help='Path to the template file')
    cmd.add_argument('-q', '--quiet', default=False, help='Only print new issue IDs after creation (for scripting)', action='store_true')

    # TODO: build template from existing issue(s)

    return parser


def main():
    parser = create_parser()
    ns = parser.parse_args()

    try:
        project = get_project(ns.project)
    except KeyError:
        sys.exit(1)
    except JIRAError as err:
        print(err)
        sys.exit(1)

    # Pass this down in namespace to callbacks
    parser.add_arg('project', project)
    try:
        rc = parser.finalize(ns)
    except JIRAError as err:
        print(err)
        sys.exit(1)
    if rc:
        ret = rc[0]
        save = rc[1]  # NOQA
    else:
        print('No command specified')
        ret = 0
        save = False  # NOQA
    sys.exit(ret)


if __name__ == '__main__':
    main()



#!/usr/bin/python3

import bz2
import copy
import json

from jirate.decor import nym


_TROLLY_CONFIG_CARD = 'META:TROLLY_CONFIG'


# WARNING WARNING WARNING - upon first creation, there's a race between the time
# trello searching finds the card.  It's 10~30 seconds.
def get_config_card(trello, board_id):
    results = trello.search.run(_TROLLY_CONFIG_CARD, idBoards=board_id, modelTypes='cards')
    if results['cards']:
        # print('Found:', results['cards'])
        return results['cards'][0]

    return None


# XXX work around the fact that json doesn't let you index by integers?
def _fix_config(my_config):
    if 'card_rev_map' not in my_config:
        return my_config

    nmap = {int(key): val for key, val in my_config['card_rev_map'].items()}
    my_config['card_rev_map'] = nmap
    return my_config


def _get_board_config(trello, config_card):
    try:
        ret = json.loads(config_card['desc'])
    except json.decoder.JSONDecodeError:
        return None

    # Not an attachment
    if 'attached' not in ret or not ret['attached']:
        return _fix_config(ret)

    attachments = trello.cards.get_attachments(config_card['id'])
    config_id = None
    for suspect in attachments:
        if not suspect['isUpload']:
            continue
        if suspect['name'] != 'trolly-config.bz2':
            continue
        config_id = suspect['id']
        break

    if not config_id:
        return None
    try:
        config_attachment = trello.cards.get_attachment(config_card['id'], config_id, max_size=(1024 * 1024 * 4))
    except json.decoder.JSONDecodeError:
        return None

    config_data = json.loads(bz2.decompress(config_attachment['data']).decode('utf-8'))
    return _fix_config(config_data)


def get_board_config(trello, board_id):
    return _get_board_config(trello, get_config_card(trello, board_id))


def _suspect(stuff, field, value, exact=False):
    # Pass 1: exact match
    for item in stuff:
        if field not in item:
            continue
        if item[field] == value:
            return item

    if exact:
        return None

    # Pass 2/3: try lowercase or nym()
    for item in stuff:
        if field not in item:
            continue
        if item[field].lower() == value.lower():
            return item
        if nym(item[field]) == nym(value):
            return item
    return None


def _search_attachments(attachments, info):
    # Check by ID
    attachment = _suspect(attachments, 'id', info, exact=True)
    if attachment:
        return attachment

    # 3-way inexact name search
    attachment = _suspect(attachments, 'name', info)
    if attachment:
        return attachment

    # OK, filename and URL
    attachment = _suspect(attachments, 'filename', info, exact=True)
    if attachment:
        return attachment

    attachment = _suspect(attachments, 'url', info, exact=True)
    if attachment:
        return attachment
    return None


class TrollyBoard(object):
    def __init__(self, trello, url, readonly=False):
        self.trello = trello

        if url.startswith('http'):
            board_id = url.rsplit('/', 1)
        else:
            board_id = url

        self._board = trello.boards.get(board_id)
        self._board_id = self._board['id']
        self._ro = readonly

        self._config_card = None
        self._config = None
        if not readonly:
            config_card = get_config_card(trello, self._board_id)
            if config_card:
                self._config_card = config_card['id']
                self._config = _get_board_config(trello, config_card)

        self.refresh()
        if not self._config_card:
            self.index_cards()

    def refresh(self):
        if not self._config:
            self._config = {'lists': {},
                            'list_map': {},
                            'default_list': None,
                            'card_map': {},
                            'card_rev_map': {}}

        self.refresh_lists()

        # Rebuild our reversemap just in case
        rev_map = {int(val): key for key, val in self._config['card_map'].items()}
        self._config['card_rev_map'] = rev_map

    def refresh_lists(self):
        lists = self.trello.boards.get_list(self._board_id)
        # XXX this shouldn't be needed; but the search ignores closed lists
        curr_lists = set([item['id'] for item in lists])
        config_lists = set(self._config['list_map'].keys())
        to_remove = list(config_lists - curr_lists)

        # If we closed a list, unlink it
        for listid in to_remove:
            # purge from our lists
            if listid in self._config['list_map']:
                # print('Unlinking list', listid)
                del self._config['lists'][self._config['list_map'][listid]]
                del self._config['list_map'][listid]

        for item in lists:
            # XXX consistency checks for the configuration?
            # print('checking', item['id'])

            if not self._config['default_list'] or self._config['default_list'] not in self._config['list_map']:
                # print('Resetting default start list to', item['id'])
                self._config['default_list'] = item['id']

            # Update in case we renamed list on the UI side
            if item['id'] in self._config['list_map']:
                self._config['lists'][self._config['list_map'][item['id']]]['name'] = item['name']
                continue

            val = {}
            val['name'] = item['name']
            val['id'] = item['id']
            name = nym(val['name'])
            while name in self._config['lists']:
                name = name + '_'
            self._config['lists'][name] = val
            self._config['list_map'][item['id']] = name

    # Unlike lists, don't use nyms for now
    def refresh_labels(self, force=True):
        if not force and 'labels' in self._config:
            return
        self._config['labels'] = self.trello.boards.get_labels(self._board_id, limit=1000)

    def labels(self):
        self.refresh_labels(False)
        return self._config['labels']

    def label_card(self, card_idx, label_name):
        card = self.card(card_idx)
        label = _suspect(card['labels'], 'name', label_name)
        if label:
            return card

        self.refresh_labels(False)
        label = _suspect(self._config['labels'], 'name', label_name)
        if label:
            return self.trello.cards.new_label_idLabel(card['id'], label['id'])
        return self.trello.cards.new_label(card['id'], label_name)

    def label_color(self, label_name, color):
        self.refresh_labels(False)
        label = _suspect(self._config['labels'], 'name', label_name)
        if label:
            return self.trello.labels.update(label['id'], color=color)
        return None

    def label_rename(self, label_name, new_name):
        self.refresh_labels(False)
        label = _suspect(self._config['labels'], 'name', label_name)
        if label:
            return self.trello.labels.update(label['id'], name=new_name)
        return None

    def refresh_members(self, force=True):
        if not force and 'labels' in self._config:
            return
        self._config['members'] = self.trello.boards.get_member(self._board_id)
        return self._config['members']

    def members(self):
        return self.refresh_members(False)

    def assign(self, card_idx, users):
        card = self.card(card_idx)
        if not card:
            return None
        if users:
            members = self.members()
            if not isinstance(users, list):
                users = [users]
            user_ids = [mb['id'] for mb in members if mb['username'] in users]
            if 'me' in users:
                user = self.trello.members.me()
                if user['id'] not in user_ids:
                    user_ids.append(user['id'])
        else:
            # Just me
            user = self.trello.members.me()
            user_ids = [user['id']]
        for user in user_ids:
            self.trello.cards.new_member(card['id'], user)

    def unassign(self, card_idx, users):
        card = self.card(card_idx)
        if not card:
            return None
        if users:
            members = self.members()
            if not isinstance(users, list):
                users = [users]
            user_ids = [mb['id'] for mb in members if mb['username'] in users]
            if 'me' in users:
                user = self.trello.members.me()
                if user not in users:
                    users.append(user)
        else:
            # Just me
            user = self.trello.members.me()
            user_ids = [user['id']]
        for user in user_ids:
            self.trello.cards.delete_member_idMember(user, card['id'])

    def unlabel_card(self, card_idx, label_name):
        card = self.card(card_idx)
        if not card:
            return None
        if 'labels' not in card:
            return card
        label = _suspect(card['labels'], 'name', label_name)
        if label:
            return self.trello.cards.delete_label_idLabel(label['id'], card['id'])
        return card

    def delete_label(self, label_name):
        self.refresh_labels(True)
        label = _suspect(self._config['labels'], 'name', label_name)
        if label:
            self.trello.labels.delete(label['id'])
        return label

    def gc_labels(self, dry_run=False):
        self.refresh_labels()
        ret = []
        for label in self._config['labels']:
            if 'name' not in label or not label['name']:
                ret.append(label)
        if not dry_run:
            for label in ret:
                self.trello.labels.delete(label['id'])
        return ret

    def list_to_id(self, list_alias):
        if list_alias not in self._config['lists'] and list_alias not in self._config['list_map']:
            raise KeyError('No such list: ' + list_alias)
        if list_alias in self._config['lists']:
            return self._config['lists'][list_alias]['id']
        return list_alias  # must be the ID

    def card_id(self, index):
        if 'card_rev_map' not in self._config:
            return None
        if index not in self._config['card_rev_map']:
            return None
        return self._config['card_rev_map'][index]

    def _index_card(self, card):
        if card['id'] not in self._config['card_map']:
            cid = card['id']
            idx = int(card['idShort'])

            # Store forward and reverse maps
            self._config['card_map'][cid] = idx
            self._config['card_rev_map'][idx] = cid

    def _index_cards(self, cards):
        if 'card_map' not in self._config:
            self._config['card_map'] = {}
        if 'card_rev_map' not in self._config:
            self._config['card_rev_map'] = {}

        for card in cards:
            self._index_card(card)

    # Prune all invisible or closed cards from our tables
    # Originally, this would look for cards which are closed, but due to how
    # Trello's filters work, it's quicker to simply scrap our tables and
    # reindex them.
    #
    # board_cleanup = 'list': prune all cards not on a current in-use list
    #               = 'all': prune all archived cards and cards as above
    def gc_cards(self, board_cleanup=None, dry_run=False):
        if board_cleanup not in (None, 'all', 'list'):
            raise ValueError('Invalid value for board_cleanup: ' + board_cleanup)
        cards = self.trello.boards.get_card_filter('visible', self._board_id)
        self._config['card_rev_map'] = {}
        self._config['card_map'] = {}
        self._index_cards(cards)

        if board_cleanup is None:
            return None

        # Completely nuke everything that is not on a visible list. This cannot
        # be undone and is destructive.  However, for long-lived boards,
        # old cards pile up and slow things down.
        ret = {}
        all_cards = self.trello.boards.get_card_filter('all', self._board_id)
        for card in all_cards:
            # We just indexed these
            if ((board_cleanup == 'all' and card['id'] in self._config['card_map']) or (board_cleanup == 'list' and card['idList'] in self._config['list_map'])):
                continue
            if card['id'] == self._config_card:
                continue
            item = {'id': card['id'], 'name': card['name']}
            ret[card['idShort']] = item
            if not dry_run:
                self.trello.cards.delete(card['id'])
        return ret

    def index_cards(self, list_alias=None):
        if list_alias is None:
            cards = self.trello.boards.get_card_filter('visible', self._board_id)
        else:
            cards = self.trello.lists.get_card(self.list_to_id(list_alias))
        self._index_cards(cards)
        return cards

    def _simplify_card_list(self, cards, userid=None):
        ret = {}
        for card in cards:
            if card['closed']:
                continue
            if card['idList'] not in self._config['list_map']:
                continue
            if userid is not None and userid not in card['idMembers']:
                continue
            val = {}
            val['id'] = card['id']
            val['name'] = card['name']
            val['list'] = self._config['list_map'][card['idList']]
            if 'labels' in card:
                val['labels'] = card['labels']
            ret[self._config['card_map'][card['id']]] = val
        return ret

    def search(self, text):
        if not text:
            return None
        ret = self.trello.search.run(text)
        return self._simplify_card_list(ret['cards'])

    def list(self, list_alias=None, userid=None):
        if userid == 'me':
            user = self.trello.members.me()
            userid = user['id']
        cards = self.index_cards(list_alias)
        return self._simplify_card_list(cards, userid)

    def card(self, card_index, verbose=False):
        need_index = False
        try:
            card_index = int(card_index)
            if card_index not in self._config['card_rev_map']:
                need_index = True
        except ValueError:
            if card_index not in self._config['card_map']:
                need_index = True
        if need_index:
            self.index_cards()

        if card_index in self._config['card_rev_map']:
            card_id = self._config['card_rev_map'][card_index]
        elif card_index in self._config['card_map']:
            card_id = card_index
        else:
            return None

        card = self.trello.cards.get(card_id)
        if verbose:
            actions = self.trello.cards.get_action(card_id, filter='all')
            card['history'] = actions
        return card

    def move(self, card_indices, list_alias):
        list_id = self.list_to_id(list_alias)
        if list_alias not in self._config['lists'] and list_alias not in self._config['list_map']:
            raise KeyError('No such list: ' + list_alias)

        if not isinstance(card_indices, list):
            card_indices = [card_indices]
        fails = []
        moves = []
        refreshed = False
        for idx in card_indices:
            try:
                idx = int(idx)
            except ValueError:
                raise ValueError('Not an index: ' + idx)
            try:
                card_id = self._config['card_rev_map'][idx]
            except KeyError:
                card_id = None

            # Only fetch once per call
            if not card_id and not refreshed:
                card = self.card(idx)
                if card:
                    card_id = card['id']
            if not card_id:
                fails.append(idx)
            else:
                moves.append(card_id)

        # Do nothing unless we can move all the cards
        if fails:
            raise ValueError('No such card(s): ' + str(fails))
        for card in moves:
            self.trello.cards.update(card, idList=list_id)
        return moves

    def link(self, index, url, text):
        card = self.card(index)
        if not card:
            raise ValueError('No such card: ' + str(index))
        return self.trello.cards.new_attachment(card['id'], url, text)

    def attach(self, index, filename):
        # Attach a physical file to a card.
        pass

    def detach(self, index, info):
        # Remove link/attachment by ID or name.
        # If ambiguous (>1 with same name)
        # raise valueerror
        card = self.card(index)
        if not card:
            raise ValueError('No such card: ' + str(index))

        attachments = self.trello.cards.get_attachments(card['id'])
        attachment = _search_attachments(attachments, info)
        if not attachment:
            return None
        return self.trello.cards.delete_attachment(attachment['id'], card['id'])

    def default_list(self, list_alias=None):
        if list_alias is None:
            return self._config['default_list']
        if list_alias not in self._config['lists'] and list_alias not in self._config['list_map']:
            raise KeyError('No such list: ' + list_alias)
        if list_alias in self._config['lists']:
            self._config['default_list'] = self._config['lists'][list_alias]['id']
        elif list_alias in self._config['list_map']:
            self._config['default_list'] = list_alias
        else:
            raise Exception('Code path error')
        return self._config['default_list']

    def rename(self, list_alias, new_name):
        if list_alias not in self._config['lists'] and list_alias not in self._config['list_map']:
            raise KeyError('No such list: ' + list_alias)

        if list_alias in self._config['lists']:
            list_id = self._config['lists'][list_alias]['id']
            list_name = list_alias
        elif list_alias in self._config['list_map']:
            list_name = self._config['list_map'][list_alias]
            list_id = list_alias

        self._config['list_map'][list_id] = new_name
        val = self._config['lists'][list_name]
        del self._config['lists'][list_name]
        self._config['lists'][new_name] = val

    def lists(self):
        return copy.copy(self._config['lists'])

    def new(self, name, description=None, start_list=None):
        if start_list is None:
            start_list = self._config['default_list']
        list_id = self.list_to_id(start_list)
        ret = self.trello.cards.new(name, list_id, description)
        self._index_card(ret)
        return ret

    def comment(self, card_idx, text):
        card = self.card(card_idx)
        return self.trello.cards.new_action_comment(card['id'], text)

    def close(self, card_idx):
        card_idx = int(card_idx)
        card_id = self._config['card_rev_map'][card_idx]
        return self.trello.cards.update_closed(card_id, True)

    def reopen(self, card_idx):
        card_idx = int(card_idx)
        if card_idx in self._config['card_rev_map']:
            card_id = self._config['card_rev_map'][card_idx]
            card = self.trello.cards.update_closed(card_id, False)
            return card

        # Search our board
        cards = self.trello.boards.get_card_filter('all', self._board_id)
        for card in cards:
            if int(card['idShort']) != card_idx:
                continue
            # Reopen card
            if card['closed']:
                self.trello.cards.update_closed(card['id'], False)
            # Send to our default list if it was in an archived list
            if card['idList'] not in self._config['list_map']:
                self.trello.cards.update(card['id'],
                                         idList=self._config['default_list'])
            return card

    def save_config(self):
        if self._ro:
            return

        # Create our config card if not present
        if not self._config_card:
            # print('Creating config card')
            list_id = list(self._config['list_map'].keys())[0]
            card = self.trello.cards.new(_TROLLY_CONFIG_CARD, list_id)
            self._config_card = card['id']
            # Hide our config card
            self.trello.cards.update_closed(card['id'], True)

        # Store config as text in description if <=15kb, otherwise store as
        # attachment
        if 'labels' in self._config:
            del self._config['labels']
        config_str = json.dumps(self._config, indent=2)
        if len(config_str) <= 15360:  # Trello limit is 16k for descriptions
            if 'attached' in self._config:
                # Sadly, need to do two dumps
                del self._config['attached']
                config_str = json.dumps(self._config, indent=2)
            self.trello.cards.update(self._config_card, desc=config_str)
            return

        if 'attached' not in self._config:
            new_desc = json.dumps({'attached': True})
            self.trello.cards.update(self._config_card, desc=new_desc)
            self._config['attached'] = True
            config_str = json.dumps(self._config, indent=2)

        config_info = config_str.encode('utf-8')

        attachments = self.trello.cards.get_attachments(self._config_card)
        old_config = None
        for suspect in attachments:
            if not suspect['isUpload']:
                continue
            if suspect['name'] != 'trolly-config.bz2':
                continue
            old_config = suspect['id']
            break

        # Upload new attachment, then purge the old one, just in case we
        # crash - better to have two (one slightly out of date) than none
        self.trello.cards.new_file_attachment(self._config_card, 'trolly-config.bz2',
                                              bindata=bz2.compress(config_info))
        if old_config:
            self.trello.cards.delete_attachment(old_config, self._config_card)

    def config(self):
        return copy.copy(self._config)

    def set_user_data(self, key, userdata):
        if key in ('default_list', 'lists', 'list_map', 'card_map', 'card_rev_map'):
            return KeyError('Reserved configuration keyword: ' + key)
        self._config['userdata']
        self.save_config()



#!/usr/bin/python

def _name_value(field_name, value):
    return {'name': value}


def _id_value(field_name, value):
    return {'id': value}


def _string(field_name, value):
    return value


_field_transmogrifiers = {
    'priority': _name_value
}


# Channelling ... Calvin
def transmogrify_input(**args):
    output = {}
    for field in args:
        value = args[field]
        if field in _field_transmogrifiers:
            output[field] = _field_transmogrifiers[field](field, value)
        else:
            output[field] = value
    return output



#!/usr/bin/python3

import json
import os


def get_config(filename=None):
    if not filename:
        try:
            config_file = open(os.path.expanduser('~/.jirate.json'))
        except FileNotFoundError:
            config_file = open(os.path.expanduser('~/.trolly.json'))
    else:
        config_file = open(os.path.expanduser(filename))
    config_data = config_file.read()
    config_file.close()
    config = json.loads(config_data)

    return config



#!/usr/bin/python3

import copy
import os

from toolchest.strutil import list_or_splitstr

from jira import JIRA, JIRAError
from jira.utils import json_loads
from jira.resources import Issue

from jirate.decor import nym
from jirate.jira_input import transmogrify_input


class Jirate(object):
    """High-level wrapper for python-jira"""
    def __init__(self, jira):
        self.jira = jira
        self._user = None

    @property
    def user(self):
        """This is the connected user (lazy-loaded)"""
        if self._user is None:
            # Get current user info and record it
            url = self.jira._get_url('myself')
            self._user = json_loads(self.jira._session.get(url))
        return self._user

    def attach(self, issue_alias, url, description):
        """Attach an external URL to an issue

        Parameters:
          issue_alias: int or string, could be JIRA Issue ID or key
          url: URL to attach (string)
          description: Description of link (string)
        """
        item = {'url': url, 'title': description}
        issue = self.issue(issue_alias)
        return self.jira.add_simple_link(issue, item)

    def search_users(self, username):
        """Wrapper for searching JIRA's user database

        Parameters:
          username: Partial username or email (string)

        Returns:
          list of jira.resources.User
        """
        # Search userlist for a username.  This is provided like this
        # so we can expand functionality later. Max is 50 by default;
        # we'll start with that
        users = self.jira.search_users(username)
        return users

    def search_issues(self, search_query):
        """Run a JQL search and assemble the results into one list

        Parameters:
          search_query: JQL query line (string)

        Returns:
          list of jira.resources.Issue
        """
        index = 0
        chunk_len = 50      # So we can detect end
        ret = []
        while True:
            issues = self.jira.search_issues(search_query, startAt=index, maxResults=chunk_len)
            if not len(issues):
                break
            ret.extend(issues)
            index = index + len(issues)
            if len(issues) < chunk_len:
                break
        return ret

    def fields(self, issue_alias):
        """Determine the fields available for an issue

        Parameters:
          issue_alias: int or string, could be JIRA Issue ID or key

        Returns:
          dict of fields
        """
        issue = self.issue(issue_alias)
        # XXX HERE THERE BE DRAGONS
        # NOT IMPLEMENTED UPSTREAM
        url = os.path.join(issue.raw['self'], 'editmeta')
        field_blob = json_loads(self.jira._session.get(url))
        return field_blob['fields']

    def get_user(self, username):
        """Determine JIRA's normalized username for someone

        Parameters:
          username: email, display name or username

        Returns:
          username (string)
        """
        if '@' not in username:
            return username

        users = self.jira.search_users(username)
        if len(users) > 1:
            raise ValueError(f'Multiple matching users for \'{username}\'')
        elif not users:
            raise ValueError(f'No matching users for \'{username}\'')

        return users[0].name

    def assign(self, issue_aliases, users=None):
        """Assign a set of issues to a user

        NOTE: Originally, the additional users were going to be set as
        watchers, which is another API call. We did this with Trello.
        Do we want to do that for JIRA, or should we have separate
        watch/unwatch?

        Parameters:
          issue_aliases: list of issue keys or IDs (list of strings)
          users: list of users (list of strings)

        Returns:
          list of issues successfully assigned to the user(s)
        """
        if isinstance(users, str):
            users = [users]
        if isinstance(issue_aliases, str):
            issue_aliases = [issue_aliases]

        for idx in issue_aliases:
            issue = self.issue(idx)
            if not issue:
                continue

            user_ids = []

            # first is assignee
            if users:
                for user in users:
                    if user == 'me':
                        user = self.user['name']
                    if user == 'none':
                        user_ids.append(None)
                    if user not in user_ids:
                        user_ids.append(self.get_user(user))
            else:
                # Just me
                user = self.user['name']
                user_ids = [user]
                # user_ids = [user['id']]

            if not len(users):
                return

            user = user_ids.pop(0)
            issue.update(assignee=user)

    def get_comment(self, issue_alias, comment_id):
        """Retrieve raw comment for an issue

        Parameters:
          issue_alias: issue key or ID (string)
          comment_id: ID of comment to retrieve

        Returns:
          jira.resources.Comment
        """
        issue = self.issue(issue_alias)
        return self.jira.comment(issue.raw['key'], comment_id)

    def comment(self, issue_alias, text):
        """Attach a new comment to an issue

        Parameters:
          issue_alias: issue key or ID (string)
          text: Text to attach as comment

        Returns:
          requests.Response
        """
        issue = self.issue(issue_alias)
        if not issue:
            return None
        # Use simple comment mode to add a comment
        url = os.path.join(issue.raw['self'], 'comment')
        return self.jira._session.post(url, data={'body': text})

    def close(self, issues):
        """Attach a new comment to an issue

        XXX this might not be a usable API and/or may be a higher
        level transition than we want. Note that 'Closed' may also
        not be correct for a given project, since all transitions
        are per-project. We could do something intelligent with
        move, or delete this API altogether.

        Parameters:
          issues: list of issue keys or IDs (list of string)

        Returns:
          requests.Response
        """
        return self.move(issues, 'Closed')

    def create(self, **args):
        """Create a new issue using key/value pairs

        Parameters:
          **args: Dictionary of key/value pairs (dict)

        Returns:
          jira.resources.Issue
        """
        # Structures for certain things need to be adjusted, because JIRA.
        # parent key is special because we do our own shortcuts.  Overwrite
        # project if we're creating a subtask
        if 'parent' in args and isinstance(args['parent'], str):
            parent_issue = self.issue(args['parent'])
            parent_key = parent_issue.raw['key']
            project = parent_issue.raw['fields']['project']['key']
            args['parent'] = {'key': parent_key}
            args['project'] = project

        # Transmogrify other fields
        new_args = transmogrify_input(**args)

        ret = self.jira.create_issue(**new_args)
        return ret

    def update_issue(self, issue_alias, **kwargs):
        """Update an issue using key/value pairs

        Parameters:
          **args: Dictionary of key/value pairs (dict)

        Returns:
          jira.resources.Issue (?)
        """
        issue = self.issue(issue_alias)
        if issue:
            return issue.update(**kwargs)

    def issue(self, issue_alias, verbose=False):
        """Retrieve an issue from JIRA

        XXX Cleanup vs. JiraProject

        Parameters:
          issue_alias: key or issue ID (string)

        Returns:
          jira.resources.Issue
        """
        if isinstance(issue_alias, Issue):
            return issue_alias
        issue_aliases = [issue_alias]
        if isinstance(issue_alias, int):
            issue_alias = str(issue_alias)
        if issue_alias.upper() != issue_alias:
            issue_aliases.append(issue_alias.upper())
        for alias in issue_aliases:
            try:
                issue = self.jira.issue(alias)
                if not issue:
                    continue
                return issue
            except JIRAError:
                pass
        return None

    def issues(self, issue_list, verbose=False):
        """Retrieve one or more issues from JIRA

        Parameters:
          issue_list: string of keys or list of keys (strings)

        Returns:
          list of jira.resources.Issue or None
        """
        if isinstance(issue_list, Issue):
            return issue_list
        issues = list_or_splitstr(issue_list)

        ret = []
        for issue in issues:
            ret.append(self.issue(issue, verbose))
        if ret:
            return ret
        return None

    def transitions(self, issue):
        """Retrieve possible next-state transitions for an issue

        Parameters:
          issue_alias: key or issue ID (string)

        Returns:
          dict - {'state': 'id', 'state2': 'id2'}
        """
        if isinstance(issue, str):
            issue = self.issue(issue)
        possible = {}
        url = os.path.join(issue.raw['self'], 'transitions')
        transitions = json_loads(self.jira._session.get(url))
        for transition in transitions['transitions']:
            possible[transition['to']['id']] = {'id': transition['id'], 'name': transition['to']['name']}
        if not possible:
            return None
        return possible

    def _find_transition(self, issue, status):
        transitions = self.transitions(issue)
        for state_id in transitions:
            if transitions[state_id]['name'] == status or nym(transitions[state_id]['name']) == status or str(state_id) == str(status):
                return transitions[state_id]['id']
        return None

    def move(self, issue_aliases, status):
        """Execute a transition to move a set of issues to the desired status

        Jira doesn't have a status you can update; you have to retrieve possible
        transitions and satisfy those requirements. Each issue has its own transition map
        according to the issue type within a given project

        Parameters:
          issue_aliases: list keys or issue IDs (list of string)
          status: Desired status

        Returns:
          list of successfully moved issues (list of string)
        """
        if not isinstance(issue_aliases, list):
            issue_aliases = [issue_aliases]

        fails = []
        moves = []
        issues = []
        for idx in issue_aliases:
            issue = self.issue(idx)
            if not issue:
                fails.append(idx)
                continue
            if idx in moves:
                continue
            # Don't double-move (if someone specified the same item twice)
            moves.append(idx)
            issues.append(issue)

        if fails:
            raise ValueError('No such issue(s): ' + str(fails))
        for issue in issues:
            transition = self._find_transition(issue, status)
            if not transition:
                continue

            # POST /rest/api/2/issue/{issueIdOrKey}/transitions
            url = os.path.join(issue.raw['self'], 'transitions')
            self.jira._session.post(url, data={'transition': {'id': transition}})
        return moves

    def link_types(self):
        """Wrapper for jira.issue_link_types()"""
        return self.jira.issue_link_types()

    def link(self, left_alias, right_alias, link_text):
        """Link two isues together using the noted link type

        Parameters:
          left_alias: left key or issue IDs (string)
          right_alias: right key or issue IDs (string)
          link_text: Desired link type

        Returns:
          ???
        """
        left = self.issue(left_alias)
        right = self.issue(right_alias)
        return self.jira.create_issue_link(link_text, left.raw['key'], right.raw['key'])

    def remote_links(self, issue_alias):
        """Obtain all remote links (URLs) attached to an issue

        Parameters:
          issue_alias: key or issue IDs (string)

        Returns:
          list of jira.resources.RemoteLink
        """
        issue = self.issue(issue_alias)
        links = self.jira.remote_links(issue.raw['id'])
        return links

    def unlink(self, left_alias, right_alias):
        """Break all links between to issues

        Parameters:
          left_alias: left key or issue IDs (string)
          right_alias: right key or issue IDs (string)

        Returns:
          count of links removed
        """
        left = self.issue(left_alias)
        right = self.issue(right_alias)

        if not right:
            extlink = self.jira.remote_link(left.raw['key'], str(right_alias))
            self.jira._session.delete(extlink.raw['self'])
            return 1

        info = left.raw['fields']

        if 'issuelinks' not in info or not info['issuelinks']:
            return 0

        right_name = right.raw['key']
        count = 0
        for link in info['issuelinks']:
            link_id = None
            if 'inwardIssue' in link and link['inwardIssue']['key'] == right_name:
                link_id = link['id']
            if 'outwardIssue' in link and link['outwardIssue']['key'] == right_name:
                link_id = link['id']
            if link_id:
                count = count + 1
                self.jira.delete_issue_link(link_id)
        return count


class JiraProject(Jirate):
    def __init__(self, jira, project, closed_status=None, readonly=False, allow_code=False, simplify=False):
        self.jira = jira
        self._ro = readonly
        self._config = None
        self._closed_status = closed_status
        self._project = self.jira.project(project)
        self._user = None
        self._issue_types = None
        self._simple = simplify
        self.custom_fields = None
        self.project_name = project
        self.allow_code = allow_code
        self.refresh()

        if self._closed_status is None:
            # guess at common closed states
            for status in ['CLOSED', 'DONE', 'RESOLVED']:
                try:
                    self.status_to_id(status)
                    self._closed_status = status
                    break
                except KeyError:
                    pass

    def refresh(self):
        if not self._config:
            self._config = {'states': {},
                            'issue_map': {},
                            'issue_rev_map': {}}

        self.refresh_lists()
        # self.index_issues()

    def refresh_lists(self):
        # DANGER DANGER - using private stuff because upstream doesn't have it
        spath = os.path.join(self._project._resource, 'statuses')
        url = self._project._get_url(spath.format(str(self._project)))
        status_info = json_loads(self.jira._session.get(url))
        status_ids = []
        statuses = []

        for task_type in status_info:
            for status in task_type['statuses']:
                if status['id'] not in status_ids:
                    status_ids.append(status['id'])
                    statuses.append(status)

        for item in statuses:
            val = {}
            val['name'] = item['name']
            val['id'] = item['id']
            name = nym(val['name'])
            while name in self._config['states']:
                name = name + '_'
            self._config['states'][name] = val

    def delete_issue_map(self):
        self._config['issue_map'] = {}

    def unlabel_issue(self, issue_alias, label_name):
        return False

    def status_to_id(self, status):
        status = nym(status)

        if status not in self._config['states']:
            raise KeyError('No such list: ' + status)
        if status in self._config['states']:
            return self._config['states'][status]['id']
        return status  # must be the ID

    def search_issues(self, text):
        # Override so we can index our return values
        if not text:
            return None
        ret = super().search_issues(text)
        self._index_issues(ret)
        return ret

    def _index_issue(self, issue):
        if issue.raw['key'] not in self._config['issue_map']:
            self._config['issue_map'][issue.raw['key']] = issue

    def _index_issues(self, issues):
        if 'issue_map' not in self._config:
            self._config['issue_map'] = {}

        for issue in issues:
            self._index_issue(issue)

    def index_issues(self, status=None):
        if status:
            status_id = self.status_to_id(status)
            open_issues = super().search_issues(f'PROJECT = {self.project_name} AND STATUS = {status_id}')
        else:
            open_issues = super().search_issues(f'PROJECT = {self.project_name} AND STATUS != {self._closed_status}')
        self._index_issues(open_issues)
        return open_issues

    def search(self, text):
        if not text:
            return None
        return self.search_issues(f'PROJECT = {self.project_name} AND STATUS != {self._closed_status} AND (text ~ "{text}")')

    def list(self, status=None, userid=None):
        if userid == 'me':
            userid = self.user['name']
        return self.index_issues(status)

    def issue(self, issue_alias, verbose=False):
        if isinstance(issue_alias, Issue):
            return issue_alias
        issue_aliases = [issue_alias]
        if issue_alias.upper() != issue_alias:
            issue_aliases.append(issue_alias.upper())
        if '-' not in issue_alias:
            issue_aliases.append(self.project_name.upper() + f'-{issue_alias}')
        for alias in issue_aliases:
            if alias in self._config['issue_map']:
                return self._config['issue_map'][alias]
            try:
                issue = self.jira.issue(alias)
                if not issue:
                    continue
                self._index_issue(issue)
                return issue
            except JIRAError:
                pass
        return None

    def states(self):
        return copy.copy(self._config['states'])

    def new(self, name, description=None, issue_type=None, parent=None):
        # Simple New creation requires understanding what the issuetypes are,
        # which vary on a per-project basis.  This is why "create" is separate
        # (and lower-level)
        if parent:
            if issue_type != 'Sub-task':
                raise ValueError('Specifying a parent only valid for Sub-task type')

        issuetypes = self.issue_types
        resolved_issue_type = None
        if issue_type is not None:
            for itype in issuetypes:
                if itype.name.upper() == issue_type.upper():
                    resolved_issue_type = itype.name
                    break
            if not resolved_issue_type:
                raise ValueError(f'No such issue type: {issue_type}')
        else:
            resolved_issue_type = issuetypes[0].name

        args = {}
        args['project'] = self.project_name
        args['summary'] = name
        args['issuetype'] = resolved_issue_type
        if parent is not None:
            args['parent'] = parent
        if description:
            args['description'] = description

        return self.create(**args)

    def create(self, **args):
        # override so we can index our value
        ret = super().create(**args)
        self._index_issue(ret)
        return ret

    def subtask(self, parent, name, description=None):
        return self.new(name, description, 'Sub-task', parent)

    @property
    def issue_types(self):
        if not self._issue_types:
            self._issue_types = self._project.issueTypes
        return self._issue_types

    # Returns a dict that JIRA should just give us.
    def issue_metadata(self, issue_type_or_id):
        itype = None
        for issuetype in self.issue_types:
            if issuetype.id == issue_type_or_id or nym(issuetype.name) == nym(issue_type_or_id):
                itype = issuetype
        if not itype:
            return None

        issue_type_id = itype.id
        fields = []
        start = 0
        chunk_len = 50
        while True:
            data = {'startAt': start, 'maxResults': chunk_len}
            new_fields = self.jira._get_json(f'issue/createmeta/{self.project_name}/issuetypes/{issue_type_id}', params=data)
            fields.extend(new_fields['values'])
            if new_fields['isLast']:
                break
            start = start + chunk_len

        field_dict = {val['fieldId']: val for val in fields}
        metadata = {'self': itype.self, 'name': itype.name, 'id': itype.id, 'description': itype.description, 'subtask': itype.subtask, 'iconUrl': itype.iconUrl, 'fields': field_dict}
        return metadata

    def config(self):
        return copy.copy(self._config)

    def get_user_data(self, key):
        if key in ('states', 'issue_map', 'issue_rev_map'):
            return KeyError('Reserved configuration keyword: ' + key)
        if key in self._config:
            return copy.copy(self._config[key])
        return None

    def set_user_data(self, key, userdata):
        if key in ('states', 'issue_map', 'issue_rev_map'):
            return KeyError('Reserved configuration keyword: ' + key)
        self._config[key] = copy.copy(userdata)


def get_jira(jconfig):
    """Wrapper to create a python-jira connection

    Parameters:
      jconfig: dict of 3 keys: url, token, proxies (optional)

    Returns:
      jira.JIRA
    """
    if 'url' not in jconfig:
        print('No JIRA URL specified')
        return None
    if 'token' not in jconfig:
        print('No JIRA token specified')
        return None
    if 'proxies' not in jconfig:
        jconfig['proxies'] = {"http": "", "https": ""}

    return JIRA(jconfig['url'], token_auth=jconfig['token'], proxies=jconfig['proxies'])



#!/usr/bin/python3

import os
import re
import sys
# import yaml

import editor

from trollo import TrelloApi

from jirate.args import ComplicatedArgs
from jirate.board import TrollyBoard
from jirate.decor import color_string, hbar_under, pretty_date, md_print
from jirate.config import get_config


def extract_bugzillas(card):
    ret = re.findall(r'://bugzilla\.redhat\.com/([0-9]+)', card['desc'])
    ret = list(set(ret) | set(re.findall(
                              r'://bugzilla\.redhat\.com/show_bug\.cgi\?id=([0-9]+)',
                              card['desc'])))
    return [int(r) for r in ret]


def bugzilla_refs(trello_list):
    ret = []
    for p in trello_list:
        ret = list(set(ret) | set(extract_bugzillas(p)))
    return ret


def trello_init(config):
    if config is not None:
        try:
            TRELLO_KEY = config['trello']['key']
            TRELLO_TOKEN = config['trello']['token']
        except IndexError:
            print('Please set your Trello key and token in your configuration file')
            exit(1)
    else:
        try:
            TRELLO_KEY = os.environ['TRELLO_KEY']
            TRELLO_TOKEN = os.environ['TRELLO_TOKEN']
        except KeyError:
            print('Please set your TRELLO_KEY and TRELLO_TOKEN variables or place them your config file')
            exit(1)

    if TRELLO_KEY is None:
        print("Please specify key in ~/.???")
        print("Visit this URL to obtain it:")
        print("  https://trello.com/app-key")
        exit(1)

    trello = TrelloApi(TRELLO_KEY)
    if not TRELLO_TOKEN:
        print("Visit this URL to get your token:")
        print('   https://trello.com/1/authorize?expiration=never&scope=read,write&response_type=token&name=shale&key=' + TRELLO_KEY)
        exit(1)

    #
    # Wow very simple
    #
    trello.set_token(TRELLO_TOKEN)

    return trello


def move(args):
    if len(args.src) == 1:
        try:
            args.board.rename(args.src[0], args.target)
            return (0, True)
        except KeyError:
            pass

    if args.board.move(args.src, args.target):
        print('Moved', args.src, 'to', args.target)
        return (0, False)
    return (1, False)


def close_cards_in_lists(args):
    for archive_list in args.target:
        list_id = args.board.list_to_id(archive_list)
        args.board.trello.lists.archive_all_cards(list_id)
    return (0, False)


def close_cards(args):
    ret = 0
    if args.list:
        return close_cards_in_lists(args)
    for card in args.target:
        if not args.board.close(card):
            ret = 1
    return (ret, False)


def reopen_card(args):
    ret = 0
    for card in args.card:
        if not args.board.reopen(card):
            ret = 1
    return (ret, False)


def print_cards_simple(cards, args=None):
    lists = {}
    for card in cards:
        clist = cards[card]['list']
        if args and args.list and clist not in args.list:
            continue
        if clist not in lists:
            lists[clist] = []
        lists[clist].append(card)

    for key in lists:
        print(key)
        for card in lists[key]:
            print('  ', card, end=' ')
            if args and args.labels:
                print_labels(cards[card], prefix='')
            print(cards[card]['name'])


def search_cards(args):
    ret = args.board.search(' '.join(args.text))
    if not ret:
        return (127, False)
    print_cards_simple(ret)
    return (0, False)


def list_cards(args):
    # check for verbose
    if args.mine:
        userid = 'me'
    else:
        userid = None

    cards = args.board.list(userid=userid)
    print_cards_simple(cards, args)
    return (0, True)


def list_lists(args):
    default = args.board.default_list()
    lists = args.board.lists()
    for lname in lists:
        if lists[lname]['id'] == default:
            print(' •', lname, lists[lname]['name'])
        else:
            print('  ', lname, lists[lname]['name'])
    return (0, False)


def set_default(args):
    default = args.board.default_list()
    new_default = args.board.default_list(args.list)
    return (0, (default != new_default))


def split_card_text(text):
    lines = text.split('\n')
    name = lines[0]
    desc = ''
    if not len(name):
        return (None, None)
    lines.pop(0)
    while len(lines) and lines[0] == '':
        lines.pop(0)
    if len(lines):
        desc = '\n'.join(lines)
    return (name, desc)


def new_card(args):
    desc = None
    if args.text:
        name = ' '.join(args.text)
    else:
        text = editor()
        name, desc = split_card_text(text)
        if name is None:
            print('Canceled')
            return (1, False)

    card = args.board.new(name, desc)
    print(card['idShort'], card['name'])
    return (0, True)


def comment(args):
    card_id = int(args.card)
    if args.text:
        text = ' '.join(args.text)
    else:
        text = editor()

    if not len(text):
        print('Canceled')
        return (0, False)

    args.board.comment(card_id, text)
    return (0, False)


def refresh(args):
    args.board.refresh()
    args.board.index_cards()
    return (0, True)


def action_null(action, arg=None):
    pass


def action_comment(action, verbose):
    data = action['data']
    if verbose:
        print(pretty_date(action['date']), '• Comment by', action['memberCreator']['username'], 'ID', action['id'])
    else:
        print(pretty_date(action['date']), '• Comment by', action['memberCreator']['username'])
    md_print(data['text'])


def display_move(action, verbose):
    if not verbose:
        return
    data = action['data']
    print(pretty_date(action['date']), '• Moved by', action['memberCreator']['username'])
    print('   ', data['listBefore']['name'], '→', data['listAfter']['name'])


def display_state(action, verbose):
    if not verbose:
        return
    data = action['data']
    if data['card']['closed']:
        print(pretty_date(action['date']), '• Closed by', action['memberCreator']['username'])
    else:
        print(pretty_date(action['date']), '• Opened by', action['memberCreator']['username'])


def display_card_update(action, verbose):
    if not verbose:
        return
    old = action['data']['old']
    new = action['data']['card']

    if 'desc' in old:
        print(pretty_date(action['date']), '• Description updated by', action['memberCreator']['username'])
        print('   ┄┉━ Old Description ━┉┄')
        md_print(old['desc'])
        print('   ┄┉━ New Description ━┉┄')
        md_print(new['desc'])
        print()
    if 'name' in old:
        print(pretty_date(action['date']), '• Summary updated by', action['memberCreator']['username'])
        print('   Old Name:', old['name'])
        print('   New Name:', new['name'])


update_map = {
    'idList': display_move,
    'closed': display_state,
    'due': action_null,           # Due date/time set
    'dueReminder': action_null,   # Due reminder set
    'pos': action_null,           # Priority change
    'name': display_card_update,  # Name updated
    'desc': display_card_update,  # Description updated
    'isTemplate': action_null     # Card is a template
}


def action_update(action, verbose):
    update_type = list(action['data']['old'].keys())[0]

    try:
        update_map[update_type](action, verbose)
    except KeyError:
        print('warning: unhandled action type:', action['type'] + ':' + update_type)


def action_create(action, verbose):
    print(pretty_date(action['date']), '• Created by', action['memberCreator']['username'])


action_map = {
    'commentCard': action_comment,
    'updateCard': action_update,
    'createCard': action_create,
    'addMemberToCard': action_null,
    'removeMemberFromCard': action_null,
    'addAttachmentToCard': action_null,
    'deleteAttachmentFromCard': action_null,
    'copyCard': action_null,
    'addChecklistToCard': action_null,
    'removeChecklistFromCard': action_null,
    'updateCheckItemStateOnCard': action_null,
    'updateCustomFieldItem': action_null
}


def display_action(action, verbose):
    try:
        action_map[action['type']](action, verbose)
    except KeyError:
        print('warning: unhandled action type:', action['type'])
        pass


def display_attachment(attachment, verbose):
    print('  ' + attachment['name'])
    if verbose:
        print('    ID:', attachment['id'])
    if attachment['isUpload']:
        if attachment['filename'] != attachment['name']:
            print('    Filename:', attachment['filename'])
    else:
        if attachment['url'] != attachment['name']:
            print('    URL:', attachment['url'])


def print_labels(card, prefix='Labels: '):
    if 'labels' in card and len(card['labels']):
        print(prefix, end='')
        for label in card['labels']:
            print(color_string(label['name'], 'white', bgcolor=label['color']), end=' ')
        print()


def print_card(board, card, verbose):
    sep = '┃'
    lsize = max(len(str(card['idShort'])), len('Assignee(s)'))

    print(str(card['idShort']).ljust(lsize), sep, card['name'])

    if verbose:
        print('ID'.ljust(lsize), sep, card['id'])
        print('URL'.ljust(lsize), sep, card['url'])

    print_labels(card, prefix='Labels'.ljust(lsize) + f' {sep} ')

    if 'idMembers' in card and card['idMembers']:
        board_members = board.members()
        print('Assignee(s)'.ljust(lsize), sep, end=' ')
        if verbose:
            print()
        for member in board_members:
            if member['id'] not in card['idMembers']:
                continue
            if verbose:
                print(' '.ljust(lsize), sep, '•', member['username'], '-', member['fullName'])
            else:
                print(member['username'], end=' ')
        if not verbose:
            print()

    if card['desc']:
        md_print(card['desc'])

    if int(card['badges']['attachments']):
        print()
        hbar_under('Attachments')
        attachments = board.trello.cards.get_attachments(card['id'])
        for attachment in attachments:
            display_attachment(attachment, verbose)

    print()
    hbar_under('Activity')

    for act in card['history']:
        display_action(act, verbose)


def cat(args):
    cards = []
    for card_idx in args.card_id:
        card = args.board.card(card_idx, True)
        if not card:
            print('No such card:', card_idx)
            return (127, False)
        cards.append(card)

    for card in cards:
        print_card(args.board, card, args.verbose)
    return (0, False)


def purge(args):
    dry_run = True

    if args.yes:
        dry_run = False

    cards = args.board.gc_cards('all', dry_run)
    labels = args.board.gc_labels(dry_run)
    if dry_run:
        if len(cards):
            for card in cards:
                print(' ', card, cards[card]['name'])
            print(f'The preceding {len(cards)} cards would be purged.')
        if len(labels):
            print(f'{len(labels)} unnamed labels would be purged.')
        if len(cards) or len(labels):
            print('Rerun with \'--yes\' to actually perform this operation.')
    else:
        if len(cards):
            print(f'Purged {len(cards)} cards')
        if len(labels):
            print(f'Purged {len(labels)} unnamed labels')

    return (0, False)


def join_card_text(name, desc):
    return name + '\n\n' + desc


def edit_card(args):
    if args.comment:
        comment_id = args.item
        comment = args.board.trello.actions.get(comment_id)
        if args.text:
            new_text = ' '.join(args.text)
        else:
            new_text = editor(comment['data']['text'])
            if not new_text:
                print('Canceled')
                return (0, False)
        if comment['data']['text'] != new_text:
            args.board.trello.actions.update(comment_id, new_text)
        else:
            print('No changes')
        return (0, False)

    card_idx = args.item
    card = args.board.card(card_idx)
    card_text = join_card_text(card['name'], card['desc'])
    if args.text:
        new_text = ' '.join(args.text)
    else:
        new_text = editor(card_text)
    if not new_text:
        print('Canceled')
        return (0, False)
    name, desc = split_card_text(new_text)
    if card['name'] != name or card['desc'] != desc:
        args.board.trello.cards.update(card['id'], name=name, desc=desc)
    else:
        print('No changes')
    return (0, False)


def labels(board, verbose):
    print('Labels:')
    labels = board.labels()
    labels_sorted = sorted(labels, key=lambda val: val['name'].lower())
    for label in labels_sorted:
        if not label['name']:
            label['name'] = 'UNNAMED'
        label_text = '  ' + color_string(label['name'], 'white', bgcolor=label['color'])
        if verbose:
            label_text = label_text + '  ' + label['id']
        print(label_text)
    return (0, False)


def label_card(args):
    verbose = args.verbose
    if not args.target and not args.new:
        return labels(args.board, verbose)

    if args.color:
        if not args.target:
            print('Changing colors requires a label.')
            return (1, False)
        if args.board.label_color(args.target[0], args.color):
            return (0, False)
        return (1, False)

    if args.rename:
        if not args.target or len(args.target) < 2:
            print('Renaming a label requires an old name and a new name.')
            return (1, False)
        if args.board.label_rename(args.target[0], args.target[1]):
            return (0, True)
        return (1, False)

    if args.remove:
        if len(args.target) < 2:
            args.board.delete_label(args.target[0])
        else:
            args.board.unlabel_card(args.target[0], args.target[1])
        return (0, False)
    args.board.label_card(args.target[0], args.target[1])
    return (0, False)


def link(args):
    card_id = args.card
    url = args.url
    if not len(args.name):
        name = url
    else:
        name = ' '.join(args.name)

    args.board.link(card_id, url, name)
    return (0, False)


def detach(args):
    card_id = args.card
    info = ' '.join(args.name)
    args.board.detach(card_id, info)
    return (0, False)


def view_card(args):
    card_id = args.card_id
    card = args.board.card(card_id)
    if not card:
        return (127, False)
    os.system('xdg-open ' + card['shortUrl'])
    return (0, False)


def members(args):
    members = args.board.members()
    for member in members:
        if args.verbose:
            uname = member['username']
            mname = member['fullName']
            mid = member['id']
            print(f'{uname} "{mname}" {mid}')
        else:
            print(member['username'])
    return (0, False)


def assign_card(args):
    args.board.assign(args.card_id, args.members)
    return (0, False)


def unassign_card(args):
    args.board.unassign(args.card_id, args.members)
    return (0, False)


def get_board(board_name=None):
    config = get_config()
    board_id = None
    readonly = False

    # Backwards compatibility
    if config is None:
        try:
            board_id = os.environ['TROLLY_BOARD']
        except KeyError:
            print('Please set a TROLLY_BOARD environment variable')
            raise

        try:
            readonly = os.environ['TROLLY_READONLY']
            if readonly in ('1', 'true', 'True', 'yes', 'Yes'):
                readonly = True
        except KeyError:
            pass
    else:
        if board_name is None:
            try:
                board_name = config['trello']['default_board']
            except IndexError:
                pass
        for board in config['trello']['boards']:
            # take first board if none specified and no default
            if board_name is None:
                board_name = board['name']
            if board_name == board['name']:
                board_id = board['id']
                readonly = False
                try:
                    readonly = board['readonly']
                    if readonly in ('1', 'true', 'True', 'yes', 'Yes'):
                        readonly = True
                except KeyError:
                    pass
                break

    board = trello_init(config)
    return TrollyBoard(board, board_id, readonly=readonly)


def create_parser():
    parser = ComplicatedArgs()

    parser.add_argument('-b', '--board', help='Use this Trello board (from config file)', default=None)

    cmd = parser.command('ls', help='List card(s)', handler=list_cards)
    cmd.add_argument('-m', '--mine', action='store_true', help='Display only cards assigned to me.')
    cmd.add_argument('-l', '--labels', action='store_true', help='Display card labels.')
    cmd.add_argument('list', nargs='*', help='Restrict to cards in these list(s)')

    cmd = parser.command('search', help='List card(s) with matching text', handler=search_cards)
    cmd.add_argument('text', nargs='*', help='Search text')

    cmd = parser.command('cat', help='Print card(s)', handler=cat)
    cmd.add_argument('-v', '--verbose', action='store_true', help='Verbose output')
    cmd.add_argument('card_id', nargs='+', help='Target card(s)')

    cmd = parser.command('view', help='Display card in browser', handler=view_card)
    cmd.add_argument('card_id', help='Target card')

    parser.command('ll', help='List lists on board', handler=list_lists)

    cmd = parser.command('label', help='Manage labels', handler=label_card)
    cmd.add_argument('-v', '--verbose', action='store_true', help='Verbose output')
    cmd.add_argument('-r', '--remove', help='Remove label from a card. If no card is specified, removes the label from the board', action='store_true', default=False)
    cmd.add_argument('--color', help='Specify a label\'s color',
                     choices=['blue', 'green', 'orange', 'purple', 'red', 'yellow', 'sky', 'lime', 'pink', 'black'])
    cmd.add_argument('--rename', action='store_true', default=False, help='Rename a label')
    cmd.add_argument('-n', '--new', help='Create a new label')
    cmd.add_argument('target', help='Target Card/Label', nargs='*')

    cmd = parser.command('assign', help='Assign card to board member(s)', handler=assign_card)
    cmd.add_argument('card_id', help='Target card')
    cmd.add_argument('members', help='Board members (if none, assign to self)', nargs='*')

    cmd = parser.command('unassign', help='Remove assignee(s) from card', handler=unassign_card)
    cmd.add_argument('card_id', help='Target card')
    cmd.add_argument('members', help='Card assignees (if none, remove only self)', nargs='*')

    cmd = parser.command('mv', help='Move card(s) or rename a list', handler=move)
    cmd.add_argument('src', metavar='card|list_name', nargs='+', help='Card IDs or list to rename')
    cmd.add_argument('target', help='Target list name')

    cmd = parser.command('new', help='Create a new card', handler=new_card)
    cmd.add_argument('text', nargs='*', help='Card title')

    cmd = parser.command('comment', help='Comment on a card', handler=comment)
    cmd.add_argument('card', help='Card to comment on')
    cmd.add_argument('text', nargs='*', help='Comment text')

    cmd = parser.command('members', help='Manipulate board members', handler=members)
    cmd.add_argument('-v', '--verbose', action='store_true', help='Verbose output')

    cmd = parser.command('edit', help='Edit card/comment text', handler=edit_card)
    cmd.add_argument('-c', '--comment', action='store_true', help='Edit a comment')
    cmd.add_argument('item', help='Item to edit (card or comment ID)')
    cmd.add_argument('text', nargs='*', help='New text')

    cmd = parser.command('close', help='Close (archive) card(s) or all cards in a list(s)', handler=close_cards)
    cmd.add_argument('-l', '--list', action='store_true', help='Act on lists')
    cmd.add_argument('target', nargs='+', help='Target card(s)/list(s)')

    cmd = parser.command('reopen', help='Reopen (send to board) card(s)', handler=reopen_card)
    cmd.add_argument('card', nargs='+', help='Card IDs reopen')

    cmd = parser.command('link', help='Add a link to a card', handler=link)
    cmd.add_argument('card', help='Target card')
    cmd.add_argument('url', help='URL to link')
    cmd.add_argument('name', nargs='*', help='Text for link')

    cmd = parser.command('detach', help='Remove attachment or link from card', handler=detach)
    cmd.add_argument('card', help='Target card')
    cmd.add_argument('name', nargs='+', help='ID, URL, filename, or name of attachment')

    cmd = parser.command('default', help='Set default list for new cards', handler=set_default)
    cmd.add_argument('list', help='List name')

    cmd = parser.command('purge', help='Remove archived cards and labels from the board', handler=purge)
    cmd.add_argument('--yes', default=False, action='store_true', help='Really do it')

    parser.command('refresh', help='Refresh board configuration', handler=refresh)

    return parser


def main():
    parser = create_parser()
    ns = parser.parse_args()

    try:
        board = get_board(ns.board)
    except KeyError:
        sys.exit(1)

    # Pass this down in namespace to callbacks
    parser.add_arg('board', board)
    rc = parser.finalize(ns)
    if rc:
        ret = rc[0]
        save = rc[1]
    else:
        print('No command specified')
        ret = 0
        save = False
    if save:
        # print('Saving...')
        board.save_config()
    sys.exit(ret)


if __name__ == '__main__':
    main()



#!/usr/bin/python3
#
# Copy/pasted from toolchest:
#   http://github.com/release-depot/toolchest

import re
import shutil

from dateutil.parser import parse
from pprint import PrettyPrinter

try:
    from rich.console import Console
    from rich.markdown import Markdown

    console = Console()
    _markdown = True
except ModuleNotFoundError:
    _markdown = False
    pass

display_color = True
HILIGHT = '[1m'
NORMAL = '[0m'

# xterm256 color maps for trello cards
COLORS = {'green':       2,     # NOQA
          'yellow':      11,    # NOQA
          'orange':      214,   # NOQA
          'red':         9,     # NOQA
          'purple':      93,    # NOQA
          'blue':        12,    # NOQA
          'blue-gray':   12,    # NOQA
          'turquoise':   80,    # NOQA
          'light green': 119,   # NOQA
          'pink':        205,   # NOQA
          'black':       0,     # NOQA
          'sky':         25,    # NOQA
          'lime':        10,    # NOQA
          'white':       15,    # NOQA
          'default':     7 }    # NOQA


def color_string(string, color=None, bgcolor=None):
    if display_color is not True:
        return string

    ret_string = ''
    fg_color = ''
    bg_color = ''
    if color and color in COLORS:
        fg_color = '[38;5;{0}m'.format(COLORS[color])
    if bgcolor and bgcolor in COLORS:
        bg_color = '[48;5;{0}m'.format(COLORS[bgcolor])

    if not fg_color and not bg_color:
        return string

    ret_string = '{0}{1}{2}[0m'.format(fg_color, bg_color, string)

    return ret_string


def jira2md(jira_text):
    # Replace code/noformat blocks with triple-backticks
    return re.sub(r'({code(:java)?}|{noformat})', '```', jira_text)


def md_print(markdown_text):
    if _markdown:
        fixed_text = jira2md(markdown_text)
        console.print(Markdown(fixed_text))
    else:
        print(markdown_text)


def pretty_date(date_str):
    date_obj = parse(date_str)
    return date_obj.astimezone().strftime('%F %T %Z')


def hbar(tl):
    if tl == 1:
        print('┄')
    elif tl == 2:
        print('┄┄')
    elif tl == 3:
        print('┄┉┄')
    else:
        print('┄┉' + '━' * (tl - 4) + '┉┄')


def hbar_over(text):
    tl = len(text)
    if not tl:
        return
    hbar(tl)
    print(text)


def hbar_under(text):
    tl = len(text)
    if not tl:
        return
    print(text)
    hbar(tl)


def nym(s):
    z = s.lower().replace(' ', '_')
    z.replace('\t', '_')
    return z


_pretty_print = PrettyPrinter(indent=4)


def pretty_print(obj):
    _pretty_print.pprint(obj)


# Print stuff in a format like so:
# field1 | field2 | field3....
#                 | field3-continued...
#                 | field3-continued-more...
# linesplit: None = trail off with '..', or separator
#            character (space, comma, etc.)
#
# vsep_print(arg1, width1, arg2, width2, ... argN)
#
vseparator = '┃'


def vsep_print(linesplit=None, *vals):
    global _termsize
    global vseparator

    sep = f' {vseparator} '

    fields = []
    widths = []
    consumed = 0
    maxwidth = 0  # Maxiumum width we printed (up to screen width)

    screen_width = shutil.get_terminal_size()[0]
    args = list(vals)

    if not args:
        return None

    while len(args) >= 2:
        val = args.pop(0)
        if not isinstance(val, str):
            val = str(val)
        fields.append(val)
        widths.append(int(args.pop(0)))
    fields.append(args.pop(0))

    #       field widths+ separators
    width = sum(widths) + (len(fields) - 1) * len(sep)
    for idx in range(0, len(widths)):
        print(fields[idx].ljust(widths[idx]) + sep, end='')
        consumed += widths[idx] + len(sep)

    last = fields.pop()
    if not isinstance(last, str):
        last = str(last)

    if len(last) <= (screen_width - width):
        print(last)
        return len(last) + width

    # Longer than remaining horizontal screen
    if not linesplit:
        print(last[:(screen_width - width - 2)] + '..')
        return screen_width

    max_chunk_len = screen_width - width
    chunks = last.split(linesplit)
    # Start on first line
    newline = False
    lsize = width - len(sep)
    consumed = 0
    while len(chunks):
        chunk = chunks.pop(0)
        while len(chunk):
            if newline is True:
                consumed = 0
                print(' ' * lsize + sep, end='')
            # Assume we'll get to the next line
            newline = True
            if len(chunk) > max_chunk_len:
                # If we've started a line, start a very long
                # max-chunk-len on new line
                if consumed > 0:
                    print()
                    newline = True
                    continue
                print(chunk[:max_chunk_len])
                newline = True
                chunk = chunk[max_chunk_len:]
                continue
            if len(chunk) > (max_chunk_len - consumed):
                # Start this chunk on new line
                print()
                newline = True
                continue
            # OK, we have space - print it
            print(chunk, end='')
            consumed += (len(chunk) + 1)
            if consumed < (max_chunk_len - 1):
                print(linesplit, end='')
                newline = False
            else:
                print()
                newline = True
            # Next chunk
            break
    # If we terminate on the edge of the screen, we already
    # print newline above, so just a check to avoid an erroneous
    # newline in output
    if consumed < max_chunk_len:
        print()
    return screen_width
